@@ -1,6 +1,7 @@
 var THREE = require('three');
 var msgpack = require('msgpack-lite');
 var dat = require('dat.gui').default; // TODO: why is .default needed?
+import {BufferGeometryUtils} from 'three/examples/jsm/utils/BufferGeometryUtils.js';
 import {OBJLoader2} from 'three/examples/jsm/loaders/OBJLoader2.js';
 import {ColladaLoader} from 'three/examples/jsm/loaders/ColladaLoader.js';
 import {MTLLoader} from 'three/examples/jsm/loaders/MTLLoader.js';
@@ -36,7 +37,7 @@ function merge_geometries(object, preserve_materials = false) {
             result.material = materials[0];
         }
     } else if (geometries.length > 1) {
-        result = THREE.BufferGeometryUtils.mergeBufferGeometries(geometries, true);
+        result = BufferGeometryUtils.mergeBufferGeometries(geometries, true);
         if (preserve_materials) {
             result.material = materials;
         }
@@ -473,7 +474,10 @@ class Animator {
         this.clock = new THREE.Clock();
         this.actions = [];
         this.playing = false;
+        this.time = 0;
+        this.time_scrubber = null;
         this.setup_capturer("png");
+        this.duration = 0;
     }
 
     setup_capturer(format) {
@@ -526,10 +530,26 @@ class Animator {
         }
     }
 
+    display_progress(time) {
+        this.time = time;
+        if (this.time_scrubber !== null) {
+            this.time_scrubber.updateDisplay();
+        }
+    }
+
+    seek(time) {
+        this.actions.forEach((action) => {
+            action.time = Math.max(0, Math.min(action._clip.duration, time));
+        });
+        this.mixer.update(0);
+        this.viewer.set_dirty();
+    }
+
     reset() {
         for (let action of this.actions) {
             action.reset();
         }
+        this.display_progress(0);
         this.mixer.update(0);
         this.setup_capturer(this.capturer.format);
         this.viewer.set_dirty();
@@ -539,6 +559,8 @@ class Animator {
         remove_folders(this.folder);
         this.mixer.stopAllAction();
         this.actions = [];
+        this.duration = 0;
+        this.display_progress(0);
         this.mixer = new THREE.AnimationMixer();
     }
 
@@ -551,6 +573,15 @@ class Animator {
         folder.add(this, "play");
         folder.add(this, "pause");
         folder.add(this, "reset");
+
+        // Note, for some reason when you call `.max()` on a slider controller it does
+        // correctly change how the slider behaves but does not change the range of values
+        // that can be entered into the text box attached to the slider. Oh well. We work
+        // around this by creating the slider with an unreasonably huge range and then calling
+        // `.min()` and `.max()` on it later.
+        this.time_scrubber = folder.add(this, "time", 0, 1e9, 0.001);
+        this.time_scrubber.onChange((value) => this.seek(value));
+
         folder.add(this.mixer, "timeScale").step(0.01).min(0);
         let recording_folder = folder.addFolder("Recording");
         recording_folder.add(this, "record");
@@ -572,6 +603,8 @@ class Animator {
             options.clampWhenFinished = true
         }
 
+        this.duration = 0;
+        this.progress = 0;
         for (let animation of animations) {
             let target = this.viewer.scene_tree.find(animation.path).object;
             let clip = this.loader.parseAnimations([animation.clip])[0];
@@ -579,7 +612,10 @@ class Animator {
             action.clampWhenFinished = options.clampWhenFinished;
             action.setLoop(options.loopMode, options.repetitions);
             this.actions.push(action);
+            this.duration = Math.max(this.duration, clip.duration);
         }
+        this.time_scrubber.min(0);
+        this.time_scrubber.max(this.duration);
         this.reset();
         if (options.play) {
             this.play();
@@ -590,6 +626,14 @@ class Animator {
         if (this.playing) {
             this.mixer.update(this.clock.getDelta());
             this.viewer.set_dirty();
+            if (this.duration != 0) {
+                let current_time = this.actions.reduce((acc, action) => {
+                    return Math.max(acc, action.time);
+                }, 0);
+                this.display_progress(current_time);
+            } else {
+                this.display_progress(0);
+            }
 
             if (this.actions.every((action) => action.paused)) {
                 this.pause();
@@ -653,8 +697,8 @@ class Viewer {
         this.dom_element.appendChild(this.renderer.domElement);
 
         this.scene = create_default_scene();
-        this.show_background();
         this.create_scene_tree();
+
         this.add_default_scene_elements();
         this.set_dirty();
 
@@ -677,8 +721,9 @@ class Viewer {
     }
 
     show_background() {
-        let top_color = [135, 206, 250]; // lightskyblue
-        let bottom_color = [25, 25, 112]; // midnightblue
+        var top_color = this.scene_tree.find(["Background"]).object.top_color;
+        var bottom_color =
+            this.scene_tree.find(["Background"]).object.bottom_color;
         this.scene.background = gradient_texture(top_color, bottom_color);
         this.set_dirty();
     }
@@ -779,11 +824,20 @@ class Viewer {
         save_folder.add(this, 'save_scene');
         save_folder.add(this, 'load_scene');
         save_folder.add(this, 'save_image');
-        let back_folder = this.gui.addFolder("Background");
-        back_folder.add(this, 'hide_background');
-        back_folder.add(this, 'show_background');
         this.animator = new Animator(this);
         this.gui.close();
+
+        this.set_property(["Background"],
+            "top_color", [135, 206, 250]); // lightskyblue
+        this.set_property(["Background"],
+            "bottom_color", [25, 25, 112]); // midnightblue
+        this.scene_tree.find(["Background"]).on_update = () => {
+            if (this.scene_tree.find(["Background"]).object.visible)
+                this.show_background();
+            else
+                this.hide_background();
+        };
+        this.show_background();
     }
 
     set_3d_pane_size(w, h) {
@@ -793,7 +847,11 @@ class Viewer {
         if (h === undefined) {
             h = window.innerHeight;
         }
-        this.camera.aspect = w / h;
+        if (this.camera.type == "OrthographicCamera") {
+            this.camera.right = this.camera.left + w*(this.camera.top - this.camera.bottom)/h;
+        } else {
+            this.camera.aspect = w / h;
+        }
         this.camera.updateProjectionMatrix();
         this.renderer.setSize(w, h);
         this.set_dirty();
@@ -860,6 +918,9 @@ class Viewer {
                 if ((obj.geometry.attributes.normal === undefined) || obj.geometry.attributes.normal.count === 0) {
                     obj.geometry.computeVertexNormals();
                 }
+            } else if (obj.type.includes("Camera")) {
+                this.set_camera(obj);
+                this.set_3d_pane_size();                
             }
             obj.castShadow = true;
             obj.receiveShadow = true;
@@ -878,6 +939,10 @@ class Viewer {
 
     set_property(path, property, value) {
         this.scene_tree.find(path).set_property(property, value);
+        if (path[0] === "Background") {
+            // The background is not an Object3d, so needs a little help.
+            this.scene_tree.find(path).on_update();
+        }
         // if (path[0] === "Cameras") {
         //     this.camera.updateProjectionMatrix();
         // }
@@ -937,6 +1002,9 @@ class Viewer {
         if (url === undefined) {
             url = `ws://${location.host}`;
         }
+        if (location.protocol == "https:") {
+            url = url.replace("ws:", "wss:");
+        }
         console.log(url);
         this.connection = new WebSocket(url);
         console.log("connection:", this.connection);
